/*
 * kgraft_patch_bsc1179877
 *
 * Fix for CVE-2020-29660 and CVE-2020-29661, bsc#1179877
 *
 *  Upstream commits:
 *  c8bcd9c5be24 ("tty: Fix ->session locking")
 *  54ffccbf053b ("tty: Fix ->pgrp locking in tiocspgrp()")
 *
 *  SLE12-SP2 and -SP3 commits:
 *  1cc3fb381cbb863d6905208cc78026f28cbc28de
 *  8ab07a963bcc577100f821a3ffdc5de6db6e2c7c
 *
 *  SLE12-SP4, SLE12-SP5, SLE15 and SLE15-SP1 commits:
 *  a59c61c4f55d549279d5d5ae24999564ab7ad8fb
 *  3de7edb960884db8906fc556a12413fbafe475dd
 *
 *  SLE15-SP2 commits:
 *  a9a2af9464b79bbdd980609fe04eb8f3a5b4233c
 *  c18ac30aac9ade995c81cef78b77a959f96a7da9
 *
 *
 *  Copyright (c) 2021 SUSE
 *  Author: Nicolai Stange <nstange@suse.de>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include "kgr_patch_bsc1179877.h"
#include "../kallsyms_relocs.h"

/* klp-ccp: from drivers/tty/tty_io.c */
#include <linux/types.h>
#include <linux/major.h>
#include <linux/errno.h>
#include <linux/signal.h>
#include <linux/fcntl.h>
#include <linux/sched.h>

/* klp-ccp: from include/linux/sched.h */
static rwlock_t (*kgre_tasklist_lock);

/* klp-ccp: from drivers/tty/tty_io.c */
#include <linux/interrupt.h>
#include <linux/tty.h>

/* klp-ccp: from include/linux/tty.h */
void kgrp_disassociate_ctty(int priv);

static int (*kgre_tty_paranoia_check)(struct tty_struct *tty, struct inode *inode,
			      const char *routine);

void kgrp___do_SAK(struct tty_struct *tty);
static void (*kgre_no_tty)(void);

static void (*kgre_tty_buffer_flush)(struct tty_struct *tty, struct tty_ldisc *ld);

long kgrp_tty_ioctl(struct file *file, unsigned int cmd, unsigned long arg);

static int (*kgre_tty_set_ldisc)(struct tty_struct *tty, int disc);

#ifdef CONFIG_AUDIT

static void (*kgre_tty_audit_tiocsti)(struct tty_struct *tty, char ch);

#else
#error "klp-ccp: non-taken branch"
#endif

/* klp-ccp: from drivers/tty/tty_io.c */
#include <linux/tty_driver.h>
#include <linux/file.h>
#include <linux/fdtable.h>
#include <linux/timer.h>
#include <linux/string.h>
#include <linux/init.h>
#include <linux/device.h>
#include <linux/wait.h>
#include <linux/bitops.h>
#include <linux/seq_file.h>
#include <linux/serial.h>
#include <linux/ratelimit.h>
#include <linux/uaccess.h>
#include <linux/kmod.h>

# define tty_debug_hangup(tty, f, args...)	do { } while (0)

static ssize_t (*kgre_redirected_tty_write)(struct file *, const char __user *,
							size_t, loff_t *);

long kgrp_tty_ioctl(struct file *file, unsigned int cmd, unsigned long arg);

static inline struct tty_struct *file_tty(struct file *file)
{
	return ((struct tty_file_private *)file->private_data)->tty;
}

const char *tty_name(const struct tty_struct *tty);

int tty_check_change(struct tty_struct *tty);

static long hung_up_tty_ioctl(struct file *file, unsigned int cmd,
		unsigned long arg)
{
	return cmd == TIOCSPGRP ? -ENOTTY : -EIO;
}

static spinlock_t (*kgre_redirect_lock);
static struct file *(*kgre_redirect);

void kgrp___proc_set_tty(struct tty_struct *tty)
{
	unsigned long flags;

	spin_lock_irqsave(&tty->ctrl_lock, flags);
	/*
	 * The session and fg pgrp references will be non-NULL if
	 * tiocsctty() is stealing the controlling tty
	 */
	put_pid(tty->session);
	put_pid(tty->pgrp);
	tty->pgrp = get_pid(task_pgrp(current));
	/*
	 * Fix CVE-2020-29660
	 *  -1 line
	 */
	tty->session = get_pid(task_session(current));
	/*
	 * Fix CVE-2020-29660
	 *  +1 line
	 */
	spin_unlock_irqrestore(&tty->ctrl_lock, flags);
	if (current->signal->tty) {
		tty_debug(tty, "current tty %s not NULL!!\n",
			  current->signal->tty->name);
		tty_kref_put(current->signal->tty);
	}
	put_pid(current->signal->tty_old_pgrp);
	current->signal->tty = tty_kref_get(tty);
	current->signal->tty_old_pgrp = NULL;
}

static void kgrr_proc_set_tty(struct tty_struct *tty)
{
	spin_lock_irq(&current->sighand->siglock);
	kgrp___proc_set_tty(tty);
	spin_unlock_irq(&current->sighand->siglock);
}

struct tty_struct *get_current_tty(void);

static void (*kgre_session_clear_tty)(struct pid *session);

static void (*kgre___tty_hangup)(struct tty_struct *tty, int exit_session);

void tty_vhangup(struct tty_struct *tty);

static void kgrr_tty_vhangup_session(struct tty_struct *tty)
{
	tty_debug_hangup(tty, "\n");
	(*kgre___tty_hangup)(tty, 1);
}

void kgrp_disassociate_ctty(int on_exit)
{
	struct tty_struct *tty;

	if (!current->signal->leader)
		return;

	tty = get_current_tty();
	if (tty) {
		if (on_exit && tty->driver->type != TTY_DRIVER_TYPE_PTY) {
			kgrr_tty_vhangup_session(tty);
		} else {
			struct pid *tty_pgrp = tty_get_pgrp(tty);
			if (tty_pgrp) {
				kill_pgrp(tty_pgrp, SIGHUP, on_exit);
				if (!on_exit)
					kill_pgrp(tty_pgrp, SIGCONT, on_exit);
				put_pid(tty_pgrp);
			}
		}
		tty_kref_put(tty);

	} else if (on_exit) {
		struct pid *old_pgrp;
		spin_lock_irq(&current->sighand->siglock);
		old_pgrp = current->signal->tty_old_pgrp;
		current->signal->tty_old_pgrp = NULL;
		spin_unlock_irq(&current->sighand->siglock);
		if (old_pgrp) {
			kill_pgrp(old_pgrp, SIGHUP, on_exit);
			kill_pgrp(old_pgrp, SIGCONT, on_exit);
			put_pid(old_pgrp);
		}
		return;
	}

	spin_lock_irq(&current->sighand->siglock);
	put_pid(current->signal->tty_old_pgrp);
	current->signal->tty_old_pgrp = NULL;

	tty = tty_kref_get(current->signal->tty);
	/*
	 * Fix CVE-2020-29660
	 *  +2 lines
	 */
	spin_unlock_irq(&current->sighand->siglock);

	if (tty) {
		unsigned long flags;

		/*
		 * Fix CVE-2020-29660
		 *  +1 line
		 */
		tty_lock(tty);
		spin_lock_irqsave(&tty->ctrl_lock, flags);
		put_pid(tty->session);
		put_pid(tty->pgrp);
		tty->session = NULL;
		tty->pgrp = NULL;
		spin_unlock_irqrestore(&tty->ctrl_lock, flags);
		/*
		 * Fix CVE-2020-29660
		 *  +1 line
		 */
		tty_unlock(tty);
		tty_kref_put(tty);
	} else
		tty_debug_hangup(tty, "no current tty\n");

	/*
	 * Fix CVE-2020-29660
	 *  -1 line
	 */
	/* Now clear signal->tty under the lock */
	read_lock(&(*kgre_tasklist_lock));
	(*kgre_session_clear_tty)(task_session(current));
	read_unlock(&(*kgre_tasklist_lock));
}

void tty_kref_put(struct tty_struct *tty);

static int kgrr_tiocsti(struct tty_struct *tty, char __user *p)
{
	char ch, mbz = 0;
	struct tty_ldisc *ld;

	if ((current->signal->tty != tty) && !capable(CAP_SYS_ADMIN))
		return -EPERM;
	if (get_user(ch, p))
		return -EFAULT;
	(*kgre_tty_audit_tiocsti)(tty, ch);
	ld = tty_ldisc_ref_wait(tty);
	if (!ld)
		return -EIO;
	if (ld->ops->receive_buf)
		ld->ops->receive_buf(tty, &ch, &mbz, 1);
	tty_ldisc_deref(ld);
	return 0;
}

static int tiocgwinsz(struct tty_struct *tty, struct winsize __user *arg)
{
	int err;

	mutex_lock(&tty->winsize_mutex);
	err = copy_to_user(arg, &tty->winsize, sizeof(*arg));
	mutex_unlock(&tty->winsize_mutex);

	return err ? -EFAULT: 0;
}

int tty_do_resize(struct tty_struct *tty, struct winsize *ws);

static int tiocswinsz(struct tty_struct *tty, struct winsize __user *arg)
{
	struct winsize tmp_ws;
	if (copy_from_user(&tmp_ws, arg, sizeof(*arg)))
		return -EFAULT;

	if (tty->ops->resize)
		return tty->ops->resize(tty, &tmp_ws);
	else
		return tty_do_resize(tty, &tmp_ws);
}

static int kgrr_tioccons(struct file *file)
{
	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;
	if (file->f_op->write == (*kgre_redirected_tty_write)) {
		struct file *f;
		spin_lock(&(*kgre_redirect_lock));
		f = (*kgre_redirect);
		(*kgre_redirect) = NULL;
		spin_unlock(&(*kgre_redirect_lock));
		if (f)
			fput(f);
		return 0;
	}
	spin_lock(&(*kgre_redirect_lock));
	if ((*kgre_redirect)) {
		spin_unlock(&(*kgre_redirect_lock));
		return -EBUSY;
	}
	(*kgre_redirect) = get_file(file);
	spin_unlock(&(*kgre_redirect_lock));
	return 0;
}

static int fionbio(struct file *file, int __user *p)
{
	int nonblock;

	if (get_user(nonblock, p))
		return -EFAULT;

	spin_lock(&file->f_lock);
	if (nonblock)
		file->f_flags |= O_NONBLOCK;
	else
		file->f_flags &= ~O_NONBLOCK;
	spin_unlock(&file->f_lock);
	return 0;
}

static int kgrr_tiocsctty(struct tty_struct *tty, struct file *file, int arg)
{
	int ret = 0;

	tty_lock(tty);
	read_lock(&(*kgre_tasklist_lock));

	if (current->signal->leader && (task_session(current) == tty->session))
		goto unlock;

	/*
	 * The process must be a session leader and
	 * not have a controlling tty already.
	 */
	if (!current->signal->leader || current->signal->tty) {
		ret = -EPERM;
		goto unlock;
	}

	if (tty->session) {
		/*
		 * This tty is already the controlling
		 * tty for another session group!
		 */
		if (arg == 1 && capable(CAP_SYS_ADMIN)) {
			/*
			 * Steal it away
			 */
			(*kgre_session_clear_tty)(tty->session);
		} else {
			ret = -EPERM;
			goto unlock;
		}
	}

	/* See the comment in tty_open(). */
	if ((file->f_mode & FMODE_READ) == 0 && !capable(CAP_SYS_ADMIN)) {
		ret = -EPERM;
		goto unlock;
	}

	kgrr_proc_set_tty(tty);
unlock:
	read_unlock(&(*kgre_tasklist_lock));
	tty_unlock(tty);
	return ret;
}

struct pid *tty_get_pgrp(struct tty_struct *tty);

static struct pid *session_of_pgrp(struct pid *pgrp)
{
	struct task_struct *p;
	struct pid *sid = NULL;

	p = pid_task(pgrp, PIDTYPE_PGID);
	if (p == NULL)
		p = pid_task(pgrp, PIDTYPE_PID);
	if (p != NULL)
		sid = task_session(p);

	return sid;
}

static int tiocgpgrp(struct tty_struct *tty, struct tty_struct *real_tty, pid_t __user *p)
{
	struct pid *pid;
	int ret;
	/*
	 * (tty == real_tty) is a cheap way of
	 * testing if the tty is NOT a master pty.
	 */
	if (tty == real_tty && current->signal->tty != real_tty)
		return -ENOTTY;
	pid = tty_get_pgrp(real_tty);
	ret =  put_user(pid_vnr(pid), p);
	put_pid(pid);
	return ret;
}

static int kgrp_tiocspgrp(struct tty_struct *tty, struct tty_struct *real_tty, pid_t __user *p)
{
	struct pid *pgrp;
	pid_t pgrp_nr;
	int retval = tty_check_change(real_tty);

	if (retval == -EIO)
		return -ENOTTY;
	if (retval)
		return retval;
	/*
	 * Fix CVE-2020-29660
	 *  -4 lines
	 */

	if (get_user(pgrp_nr, p))
		return -EFAULT;
	if (pgrp_nr < 0)
		return -EINVAL;
	/*
	 * Fix CVE-2020-29660
	 *  +8 lines
	 */
	spin_lock_irq(&real_tty->ctrl_lock);
	if (!current->signal->tty ||
	    (current->signal->tty != real_tty) ||
	    (real_tty->session != task_session(current))) {
		retval = -ENOTTY;
		goto out_unlock_ctrl;
	}

	rcu_read_lock();
	pgrp = find_vpid(pgrp_nr);
	retval = -ESRCH;
	if (!pgrp)
		goto out_unlock;
	retval = -EPERM;
	if (session_of_pgrp(pgrp) != task_session(current))
		goto out_unlock;
	retval = 0;
	/*
	 * Fix CVE-2020-29660, CVE-2020-29661
	 *  -1 line
	 */
	put_pid(real_tty->pgrp);
	real_tty->pgrp = get_pid(pgrp);
	/*
	 * Fix CVE-2020-29660, CVE-2020-29661
	 *  -1 line
	 */
out_unlock:
	rcu_read_unlock();
	/*
	 * Fix CVE-2020-29660
	 *  +2 lines
	 */
out_unlock_ctrl:
	spin_unlock_irq(&real_tty->ctrl_lock);
	return retval;
}

static int kgrp_tiocgsid(struct tty_struct *tty, struct tty_struct *real_tty, pid_t __user *p)
{
	/*
	 * Fix CVE-2020-29660
	 *  +3 lines
	 */
	unsigned long flags;
	pid_t sid;

	/*
	 * (tty == real_tty) is a cheap way of
	 * testing if the tty is NOT a master pty.
	*/
	if (tty == real_tty && current->signal->tty != real_tty)
		return -ENOTTY;
	/*
	 * Fix CVE-2020-29660
	 *  +1 line
	 */
	spin_lock_irqsave(&real_tty->ctrl_lock, flags);
	if (!real_tty->session)
		/*
		 * Fix CVE-2020-29660
		 *  -1 line, +1 line
		 */
		goto err;
	/*
	 * Fix CVE-2020-29660
	 *  -1 line, +8 lines
	 */
	sid = pid_vnr(real_tty->session);
	spin_unlock_irqrestore(&real_tty->ctrl_lock, flags);

	return put_user(sid, p);

err:
	spin_unlock_irqrestore(&real_tty->ctrl_lock, flags);
	return -ENOTTY;
}

static int kgrr_tiocsetd(struct tty_struct *tty, int __user *p)
{
	int disc;
	int ret;

	if (get_user(disc, p))
		return -EFAULT;

	ret = (*kgre_tty_set_ldisc)(tty, disc);

	return ret;
}

static int tiocgetd(struct tty_struct *tty, int __user *p)
{
	struct tty_ldisc *ld;
	int ret;

	ld = tty_ldisc_ref_wait(tty);
	if (!ld)
		return -EIO;
	ret = put_user(ld->ops->num, p);
	tty_ldisc_deref(ld);
	return ret;
}

static int (*kgre_send_break)(struct tty_struct *tty, unsigned int duration);

static int tty_tiocmget(struct tty_struct *tty, int __user *p)
{
	int retval = -EINVAL;

	if (tty->ops->tiocmget) {
		retval = tty->ops->tiocmget(tty);

		if (retval >= 0)
			retval = put_user(retval, p);
	}
	return retval;
}

static int tty_tiocmset(struct tty_struct *tty, unsigned int cmd,
	     unsigned __user *p)
{
	int retval;
	unsigned int set, clear, val;

	if (tty->ops->tiocmset == NULL)
		return -EINVAL;

	retval = get_user(val, p);
	if (retval)
		return retval;
	set = clear = 0;
	switch (cmd) {
	case TIOCMBIS:
		set = val;
		break;
	case TIOCMBIC:
		clear = val;
		break;
	case TIOCMSET:
		set = val;
		clear = ~val;
		break;
	}
	set &= TIOCM_DTR|TIOCM_RTS|TIOCM_OUT1|TIOCM_OUT2|TIOCM_LOOP;
	clear &= TIOCM_DTR|TIOCM_RTS|TIOCM_OUT1|TIOCM_OUT2|TIOCM_LOOP;
	return tty->ops->tiocmset(tty, set, clear);
}

static int tty_tiocgicount(struct tty_struct *tty, void __user *arg)
{
	int retval = -EINVAL;
	struct serial_icounter_struct icount;
	memset(&icount, 0, sizeof(icount));
	if (tty->ops->get_icount)
		retval = tty->ops->get_icount(tty, &icount);
	if (retval != 0)
		return retval;
	if (copy_to_user(arg, &icount, sizeof(icount)))
		return -EFAULT;
	return 0;
}

static void tty_warn_deprecated_flags(struct serial_struct __user *ss)
{
	static DEFINE_RATELIMIT_STATE(depr_flags,
			DEFAULT_RATELIMIT_INTERVAL,
			DEFAULT_RATELIMIT_BURST);
	char comm[TASK_COMM_LEN];
	int flags;

	if (get_user(flags, &ss->flags))
		return;

	flags &= ASYNC_DEPRECATED;

	if (flags && __ratelimit(&depr_flags))
		pr_warning("%s: '%s' is using deprecated serial flags (with no effect): %.8x\n",
				__func__, get_task_comm(comm, current), flags);
}

static struct tty_struct *tty_pair_get_tty(struct tty_struct *tty)
{
	if (tty->driver->type == TTY_DRIVER_TYPE_PTY &&
	    tty->driver->subtype == PTY_TYPE_MASTER)
		tty = tty->link;
	return tty;
}

long kgrp_tty_ioctl(struct file *file, unsigned int cmd, unsigned long arg)
{
	struct tty_struct *tty = file_tty(file);
	struct tty_struct *real_tty;
	void __user *p = (void __user *)arg;
	int retval;
	struct tty_ldisc *ld;

	if ((*kgre_tty_paranoia_check)(tty, file_inode(file), "tty_ioctl"))
		return -EINVAL;

	real_tty = tty_pair_get_tty(tty);

	/*
	 * Factor out some common prep work
	 */
	switch (cmd) {
	case TIOCSETD:
	case TIOCSBRK:
	case TIOCCBRK:
	case TCSBRK:
	case TCSBRKP:
		retval = tty_check_change(tty);
		if (retval)
			return retval;
		if (cmd != TIOCCBRK) {
			tty_wait_until_sent(tty, 0);
			if (signal_pending(current))
				return -EINTR;
		}
		break;
	}

	/*
	 *	Now do the stuff.
	 */
	switch (cmd) {
	case TIOCSTI:
		return kgrr_tiocsti(tty, p);
	case TIOCGWINSZ:
		return tiocgwinsz(real_tty, p);
	case TIOCSWINSZ:
		return tiocswinsz(real_tty, p);
	case TIOCCONS:
		return real_tty != tty ? -EINVAL : kgrr_tioccons(file);
	case FIONBIO:
		return fionbio(file, p);
	case TIOCEXCL:
		set_bit(TTY_EXCLUSIVE, &tty->flags);
		return 0;
	case TIOCNXCL:
		clear_bit(TTY_EXCLUSIVE, &tty->flags);
		return 0;
	case TIOCGEXCL:
	{
		int excl = test_bit(TTY_EXCLUSIVE, &tty->flags);
		return put_user(excl, (int __user *)p);
	}
	case TIOCNOTTY:
		if (current->signal->tty != tty)
			return -ENOTTY;
		(*kgre_no_tty)();
		return 0;
	case TIOCSCTTY:
		return kgrr_tiocsctty(tty, file, arg);
	case TIOCGPGRP:
		return tiocgpgrp(tty, real_tty, p);
	case TIOCSPGRP:
		return kgrp_tiocspgrp(tty, real_tty, p);
	case TIOCGSID:
		return kgrp_tiocgsid(tty, real_tty, p);
	case TIOCGETD:
		return tiocgetd(tty, p);
	case TIOCSETD:
		return kgrr_tiocsetd(tty, p);
	case TIOCVHANGUP:
		if (!capable(CAP_SYS_ADMIN))
			return -EPERM;
		tty_vhangup(tty);
		return 0;
	case TIOCGDEV:
	{
		unsigned int ret = new_encode_dev(tty_devnum(real_tty));
		return put_user(ret, (unsigned int __user *)p);
	}
	/*
	 * Break handling
	 */
	case TIOCSBRK:	/* Turn break on, unconditionally */
		if (tty->ops->break_ctl)
			return tty->ops->break_ctl(tty, -1);
		return 0;
	case TIOCCBRK:	/* Turn break off, unconditionally */
		if (tty->ops->break_ctl)
			return tty->ops->break_ctl(tty, 0);
		return 0;
	case TCSBRK:   /* SVID version: non-zero arg --> no break */
		/* non-zero arg means wait for all output data
		 * to be sent (performed above) but don't send break.
		 * This is used by the tcdrain() termios function.
		 */
		if (!arg)
			return (*kgre_send_break)(tty, 250);
		return 0;
	case TCSBRKP:	/* support for POSIX tcsendbreak() */
		return (*kgre_send_break)(tty, arg ? arg*100 : 250);

	case TIOCMGET:
		return tty_tiocmget(tty, p);
	case TIOCMSET:
	case TIOCMBIC:
	case TIOCMBIS:
		return tty_tiocmset(tty, cmd, p);
	case TIOCGICOUNT:
		retval = tty_tiocgicount(tty, p);
		/* For the moment allow fall through to the old method */
        	if (retval != -EINVAL)
			return retval;
		break;
	case TCFLSH:
		switch (arg) {
		case TCIFLUSH:
		case TCIOFLUSH:
		/* flush tty buffer and allow ldisc to process ioctl */
			(*kgre_tty_buffer_flush)(tty, NULL);
			break;
		}
		break;
	case TIOCSSERIAL:
		tty_warn_deprecated_flags(p);
		break;
	}
	if (tty->ops->ioctl) {
		retval = tty->ops->ioctl(tty, cmd, arg);
		if (retval != -ENOIOCTLCMD)
			return retval;
	}
	ld = tty_ldisc_ref_wait(tty);
	if (!ld)
		return hung_up_tty_ioctl(file, cmd, arg);
	retval = -EINVAL;
	if (ld->ops->ioctl) {
		retval = ld->ops->ioctl(tty, file, cmd, arg);
		if (retval == -ENOIOCTLCMD)
			retval = -ENOTTY;
	}
	tty_ldisc_deref(ld);
	return retval;
}

static int (*kgre_this_tty)(const void *t, struct file *file, unsigned fd);

void kgrp___do_SAK(struct tty_struct *tty)
{
#ifdef TTY_SOFT_SAK
#error "klp-ccp: non-taken branch"
#else
	struct task_struct *g, *p;
	struct pid *session;
	int		i;
	/*
	 * Fix CVE-2020-29660
	 *  +1 line
	 */
	unsigned long flags;

	if (!tty)
		return;
	/*
	 * Fix CVE-2020-29660
	 *  -1 line, +3 lines
	 */
	spin_lock_irqsave(&tty->ctrl_lock, flags);
	session = get_pid(tty->session);
	spin_unlock_irqrestore(&tty->ctrl_lock, flags);

	tty_ldisc_flush(tty);

	tty_driver_flush_buffer(tty);

	read_lock(&(*kgre_tasklist_lock));
	/* Kill the entire session */
	do_each_pid_task(session, PIDTYPE_SID, p) {
		printk(KERN_NOTICE "SAK: killed process %d"
			" (%s): task_session(p)==tty->session\n",
			task_pid_nr(p), p->comm);
		send_sig(SIGKILL, p, 1);
	} while_each_pid_task(session, PIDTYPE_SID, p);
	/* Now kill any processes that happen to have the
	 * tty open.
	 */
	do_each_thread(g, p) {
		if (p->signal->tty == tty) {
			printk(KERN_NOTICE "SAK: killed process %d"
			    " (%s): task_session(p)==tty->session\n",
			    task_pid_nr(p), p->comm);
			send_sig(SIGKILL, p, 1);
			continue;
		}
		task_lock(p);
		i = iterate_fd(p->files, 0, (*kgre_this_tty), tty);
		if (i != 0) {
			printk(KERN_NOTICE "SAK: killed process %d"
			    " (%s): fd#%d opened to the tty\n",
				    task_pid_nr(p), p->comm, i - 1);
			force_sig(SIGKILL, p);
		}
		task_unlock(p);
	} while_each_thread(g, p);
	read_unlock(&(*kgre_tasklist_lock));
	/*
	 * Fix CVE-2020-29660
	 *  +1 line
	 */
	put_pid(session);
#endif
}



static struct kgr_kallsyms_reloc kgr_funcs[] = {
	{ "tasklist_lock", (void *)&kgre_tasklist_lock },
	{ "redirect", (void *)&kgre_redirect },
	{ "redirect_lock", (void *)&kgre_redirect_lock },
	{ "tty_paranoia_check", (void *)&kgre_tty_paranoia_check },
	{ "no_tty", (void *)&kgre_no_tty },
	{ "tty_buffer_flush", (void *)&kgre_tty_buffer_flush },
	{ "tty_set_ldisc", (void *)&kgre_tty_set_ldisc },
	{ "tty_audit_tiocsti", (void *)&kgre_tty_audit_tiocsti },
	{ "redirected_tty_write", (void *)&kgre_redirected_tty_write },
	{ "__tty_hangup", (void *)&kgre___tty_hangup },
	{ "session_clear_tty", (void *)&kgre_session_clear_tty },
	{ "send_break", (void *)&kgre_send_break },
	{ "this_tty", (void *)&kgre_this_tty },
};

int kgr_patch_bsc1179877_init(void)
{
	return __kgr_resolve_kallsyms_relocs(kgr_funcs, ARRAY_SIZE(kgr_funcs));
}
